/* module pdfg.cpp                                               */
/* Writes a very simple PDF document                             */
/*   generates a document info                                   */
/*   generates a text annotation                                 */
/*   generates a link link annotation                            */
/*   generates ein B&W bitmap image from an external file        */
/*   demosntrates the use of /OpenAction                         */
/* Copyright Dr.E.Huckert 1-96                                   */

/* Note: this is a quick-and-dirty program                         */
/*       I sell two portable C++ classes with sample applications: */
/*       a text-to-PDF conversion class with some                  */
/*         formatting capabilities                                 */
/*       a TIFF-to-PDF conversion class for B&W images             */
/*         (including multi-part TIFF and conversion of complete   */
/*          directories)                                           */
/* For details contact huckert@compuserve.com                      */

// versions
// 17.05.98 added an /OpenAction in the root object to start with
//          the second sample page

#include <stdio.h>
#include <fcntl.h>
#include <string.h>

/* may be necessary - depending on compiler */
/* #include <unistd.h> */

#ifdef LINUX
#include <linux/iso_fs.h>
/* strange ... */
#define O_BINARY ISOFS_FILE_BINARY
/* wo ist unter LINUX die richtige Bibliothek?... */
#define setmode(x,y)
#define CREAT_MASK  0755
#else
#include <io.h>
#define CREAT_MASK 0xff        /* for MSDOS only */
#endif

#define MAXOBJS     100

extern "C"
{
  int exit(int);
};

class PDF
{
private:
  int ofd;
  int no_objs;
  int act_no;
  int root_objno;
  int info_objno;
  long objpos[MAXOBJS];
  long xrefpos;
  long writepos;
  long imglength;
public:
  PDF(char *filename);
  ~PDF();  
  int StartObject(int no);
  int StartObject(int no, char *type);
  int CloseObject();
  int AddLine(char *line);
  int AddDict(char *line);
  int StartDict();
  int CloseDict();
  int MakeRootObject(int objno, int pages,
                     int outlines, int openaction);
  int MakeAnnotation(int objno, int x1, int y1, 
                     int x2, int y2, char *txt);
  int MakeLinkAnnotation(int objno, int link_to, 
                         int x1, int y1, 
                         int x2, int y2);
  int MakeExtLinkAnnotation(int objno, char *link_to, 
                            int x1, int y1, 
                            int x2, int y2);
  int MakeInfo(int objno, char *author, char *creator, char *date);
  int MakeTextContent(int objno, char *txt);
  int MakeFont(int objno, int fontno, char *fontname);
  long WriteBmp(char *filename);
  int MakeImage(int objno, char *fn, int xdim, int ydim, int zdim);
  int MakeOutline(int objno, int parent, int next, int dest, char *title);
};

// constructor
PDF::PDF(char *filename)
{
int n;
  no_objs    = 0;
  ofd        = 0;
  act_no     = 0;
  info_objno = 0;
  root_objno = 0;
  xrefpos    = 0L;
  writepos   = 0L;
  for (n=0; n < MAXOBJS; n++)
    objpos[n] = 0L;
  ofd = creat(filename, CREAT_MASK);
  if (ofd < 0)
  {
    fprintf(stderr, "Error:cannot create file\n");
    exit(2);
  }
  setmode(ofd, O_BINARY);
  this->AddLine("%PDF-1.1\n");  
}  // end PDF::PDF()

// Destructor
// Generates cross reference and trailer
PDF::~PDF()
{
int n;
char hbuf[128];
  if (ofd != 0)
  {
    /* generate a cross reference   */
    this->xrefpos = this->writepos;
    sprintf(hbuf, "xref\n0 %d\n", this->no_objs + 1);
    this->AddLine(hbuf);
    /* a line must have exactly 20 chars incl. the linefeed */
    this->AddLine("0000000000 65535 f \n");
    for (n=1; n <= no_objs; n++)
    {
      sprintf(hbuf, "%010ld 00000 n \n", objpos[n]);
      this->AddLine(hbuf);
    }
    /* generate a trailer  */
    this->AddLine("trailer\n<<\n");
    sprintf(hbuf, "/Size %d\n", this->no_objs + 1);
    this->AddLine(hbuf);
    if (this->root_objno != 0)
    {
      sprintf(hbuf, "/Root %d 0 R\n", root_objno);
      this->AddLine(hbuf);
    }
    if (this->info_objno != 0)
    {
      sprintf(hbuf, "/Info %d 0 R\n", info_objno);
      this->AddLine(hbuf);
    }
    this->CloseDict();
    sprintf(hbuf,"startxref\n%ld\n", this->xrefpos);
    this->AddLine(hbuf);
    write(this->ofd,"%%EOF\n", 6);
    close(ofd);
  }
}   // end PDF::~PDF()

int PDF::StartObject(int no)
{
char hbuf[40];
int  nb;
  this->act_no = no;
  if (no < MAXOBJS)
    this->objpos[this->act_no] = this->writepos;
  sprintf(hbuf, "%d 0 obj\n", no);
  nb = this->AddLine(hbuf);
  if (no > this->no_objs)
    this->no_objs = no;
  return nb;
}   // end PDF::StartObject()

// book keeping for some important object ypes
int PDF::StartObject(int no, char *type)
{
int nb;
  nb = StartObject(no);
  if (strcmp(type, "/Root") == 0)
    root_objno = no;
  if (strcmp(type, "/Info") == 0)
    info_objno = no;
  return nb;
}   // end PDF::StartObject()

// write a sequence of bytes to an object
int PDF::AddLine(char *str)
{
int nb;
  nb = write(this->ofd, str, strlen(str));
  this->writepos += nb;
  return nb;
}

// start a PDF dictionary
int PDF::StartDict()
{
int nb;
  nb = this->AddLine("<<\n");
  return nb;
}

// close a PDF dictionary
int PDF::CloseDict()
{
int nb;
  nb = this->AddLine(">>\n");
  return nb;
}

// opens and closes a PDF dictionary
int PDF::AddDict(char *str)
{
int nb;
  this->StartDict();
  this->AddLine(str);
  this->CloseDict();
  return nb;
}

// closes a PDF object
int PDF::CloseObject()
{
int nb;
  nb = this->AddLine("endobj\n");
  return nb;
}

// Generate a root object
// An open action can be added if desired
int PDF::MakeRootObject(int objno,
                        int pages_objno,
                        int outlines_objno,
                        int openaction_objno) // <= 0 if no action desired
{
char buf[256];
char buf2[128];
  this->StartObject(objno, "/Root");
  sprintf(buf, "/Type /Catalog\n/Pages %d 0 R\n/Outlines %d 0 R\n",
          pages_objno, outlines_objno);
  if (openaction_objno > 0)
  {
    sprintf(buf2, "/OpenAction [%d 0 R /Fit]\n", openaction_objno);
    strcat(buf, buf2);
  }
  this->AddDict(buf);
  this->CloseObject();
  return 0;
}   // end PDF::MakeRootObject()

// Generate an annotation object
int PDF::MakeAnnotation(int objno,
                        int x1, int y1, int x2, int y2,
                        char *txt)
{
char hbuf[256];
int nb;
  this->StartObject(objno);
  this->StartDict();
  this->AddLine("/Type /Annot /Subtype /Text\n");
  sprintf(hbuf, "/Rect [%d %d %d %d]\n", x1, y1, x2, y2);
  this->AddLine(hbuf);
  this->AddLine("/Contents ");
  if (txt != NULL)
  {
    sprintf(hbuf, "(%s)\n", txt);
    this->AddLine(hbuf);
  }
  this->CloseDict();
  nb = this->CloseObject();
  return nb;
}   

// Generate an internal link (to a page in the same file)
int PDF::MakeLinkAnnotation(int objno, int link_to,
                        int x1, int y1, int x2, int y2)
{
char hbuf[256];
int nb;
  this->StartObject(objno);
  this->StartDict();
  this->AddLine("/Type /Annot /Subtype /Link\n");
  sprintf(hbuf, "/Rect [%d %d %d %d]\n", x1, y1, x2, y2);
  this->AddLine(hbuf);
  this->AddLine("/C [1 0 0]\n");    // red rectangle as link symbol
  this->AddLine("/Border [0 0 2]\n"); 
  sprintf(hbuf, "/Dest [%d 0 R /XYZ 0 0 0]\n", link_to);
  this->AddLine(hbuf);
  this->CloseDict();
  nb = this->CloseObject();
  return nb;
}   

// generate a link to an external file
int PDF::MakeExtLinkAnnotation(int objno, char *link_to,
                               int x1, int y1, int x2, int y2)
{
char hbuf[256];
int nb;
  this->StartObject(objno);
  this->StartDict();
  this->AddLine("/Type /Annot\n/Subtype /Link\n");
  sprintf(hbuf, "/Rect [%d %d %d %d]\n", x1, y1, x2, y2);
  this->AddLine(hbuf);
  this->AddLine("/C [1 0 0]\n");    // red rectangle as link symbol
  this->AddLine("/Border [0 0 2]\n"); 
  this->AddLine("/Title (Externer Verweis)\n");
  sprintf(hbuf, 
  "/A << /Type /Action /S /GoToR /D [0 /FitB] /F (%s) >>\n", link_to);
  this->AddLine(hbuf);
  this->CloseDict();
  nb = this->CloseObject();
  return nb;
}   // end PDF::MakeExtLinkAnnotation()

// creates a PDF outline object
int PDF::MakeOutline(int objno, int parent, 
                     int next, int dest, 
                     char *title)
{
char hbuf[256];
int nb;
  this->StartObject(objno);
  this->StartDict();
  sprintf(hbuf, "/Parent %d 0 R\n", parent);
  this->AddLine(hbuf);
  sprintf(hbuf, "/Dest [%d 0 R /XYZ 0 800 0]\n", dest);
  this->AddLine(hbuf);
  sprintf(hbuf, "/Next %d 0 R\n", next);
  this->AddLine(hbuf);
  sprintf(hbuf, "/Count 0\n");
  this->AddLine(hbuf);
  sprintf(hbuf, "/Title (%s)\n", title);
  this->AddLine(hbuf);
  this->CloseDict();
  nb = this->CloseObject();
  return nb;
}   

/* Generates an Info object */
int PDF::MakeInfo(int objno, char *author, char *creator, char *date)
{
int  nb;
char hbuf[256];
  this->StartObject(objno, "/Info");
  this->StartDict();
  if (creator != NULL)
  {
    sprintf(hbuf, "/Creator (%s)\n", creator);
    this->AddLine(hbuf);
  }
  if (author != NULL)
  {
    sprintf(hbuf, "/Author (%s)\n", author);
    this->AddLine(hbuf);
  }
  if (date != NULL)
  {
    sprintf(hbuf, "/CreationDate (%s)\n", date);
    this->AddLine(hbuf);
  }
  this->CloseDict();
  nb = this->CloseObject();
  return nb;
}

/* Generates a text object (/Content) */
int PDF::MakeTextContent(int objno, 
                         char *txt)
{
int  nb;
char hbuf[128];
  this->StartObject(objno);
  sprintf(hbuf, "<< /Length %d >>\nstream\n", strlen(txt));
  this->AddLine(hbuf);
  this->AddLine(txt);
  this->AddLine("endstream\n");
  nb = this->CloseObject();
  return nb;
}

/* generates a font object */  
int PDF::MakeFont(int objno, int fontno, char *fontname)
{
int nb;
char hbuf[128];
  this->StartObject(objno);
  this->StartDict();
  this->AddLine("/Type /Font\n");
  this->AddLine("/Subtype /Type1\n");
  sprintf(hbuf, "/Name /F%d\n", fontno);
  this->AddLine(hbuf);
  sprintf(hbuf, "/BaseFont /%s\n", fontname);
  this->AddLine(hbuf);
  this->AddLine("/Encoding /WinAnsiEncoding\n");
  this->CloseDict();
  this->CloseObject();
  return nb;
}

// read a binary bitmap data file (no header etc.)
// write the data hex encoded
// Returns: the length of the encoded stream
long PDF::WriteBmp(char *fn)
{
int  m, n, src;
long length;
char rbuf[20];
char buf[80];
  length = 0L;
  src = open(fn, O_RDONLY);
  if (src < 0)
    return -1L;
  setmode(src, O_BINARY);
  while (1)
  {
    n = read(src, rbuf, sizeof(rbuf));  
    if (n <= 0) break;
    for (m = 0; m < n; m++)
      sprintf(&buf[m*2], "%02x", rbuf[m] & 0xff);
    buf[n * 2] = 0;
    strcat(buf, "\n");
    this->AddLine(buf);
    length += strlen(buf);
  }
  // add an EOD character
  this->AddLine(">\n");
  length += 2;
  close(src);
  return length;
}   /* end PDF::WriteBmp() */

// only for bitmap files(1 bit/pixel), no header
// the bitmap data are written in ASCIIHexDecode filter
int PDF::MakeImage(int objno, 
                   char *filename, // name of the bitmap file
                   int xdim, int ydim, int zdim)
{
int nb;
char hbuf[128];
  this->imglength = 0L;
  this->StartObject(objno);
  this->StartDict();
  this->AddLine("/Type /Xobject\n");
  this->AddLine("/Subtype /Image\n");
  this->AddLine("/Name /bild1\n");
  sprintf(hbuf, "/Width %d /Height %d /BitsPerComponent %d\n",
          xdim, ydim, zdim);
  this->AddLine(hbuf);
  sprintf(hbuf, "/Length %d 0 R\n", this->act_no + 1);
  this->AddLine(hbuf);
  this->AddLine("/ColorSpace /DeviceGray\n"); 
  this->AddLine("/Filter /ASCIIHexDecode\n");
  //sprintf(hbuf, "/Columns %d\n/Rows %d\n", xdim, ydim);
  //this->AddLine(hbuf);
  //
  // add the bitmap data (hex notation)
  nb = this->AddLine(">>\nstream\n");
  this->imglength = this->WriteBmp(filename);
  //
  // close the image stream object
  this->AddLine("endstream\n>>\n");
  this->CloseObject();
  //
  // write a special length object for the image
  this->StartObject(this->act_no + 1);
  sprintf(hbuf, "%ld\n", this->imglength);
  this->AddLine(hbuf);
  nb = this->CloseObject();
  return nb;
}

char text1[] = "BT\n/F1 24 Tf \
1 0 0 1 100 740 Tm (Fuer den Verweis ist Datei yy.pdf noetig) Tj \
1 0 0 1 100 720 Tm (You need a file yy.pdf for the external link) Tj \
1 0 0 1 100 700 Tm [(text1)-500(text2)-500(text3)] TJ \
1 0 0 1 100 680 Tm [(text1)-999(text2)-999(text3)] TJ \
1 0 0 1 150 660 Tm (Verschobener Text)Tj \
1.8 0 0 1.8 160 620 Tm (Skalierter Text) Tj \
1 -0.1 1.5 1 300 700 Tm (Gedrehter Text) Tj ET\n\
6 w 100 600 m\n\
400 600 l 400 400 l 100 400 l 100 600 l S\n";

char text2[] = 
"BT\n/F1 24 Tf\n100 700 Td 1 0 1 rg (gefaerbter Text fuer Seite 2)Tj\nET\n\
6 w 100 600 m\n400 600 l 400 400 l 100 400 l 100 600 l S\n";

main(argc, argv)
int argc;
char *argv[];
{
int  ofd, n;
char hbuf[256];
char ibuf[128];
PDF  *pdf;
  if (argc <= 1)
  {
     fprintf(stderr,"usage:pdfg outfile\n");
     exit(1);
  }
  pdf = new PDF(argv[1]);

  // Note: the root object containing /Catalog is writte later!

  /* Write a /Pages Object, No. 2 */
  pdf->StartObject(2);
  pdf->AddDict("/Type /Pages\n/Count 2\n/Kids [3 0 R 12 0 R]\n");
  pdf->CloseObject();

  /* Write a /Page Object, No. 3 */
  pdf->StartObject(3);
  pdf->StartDict();
  pdf->AddLine("/Type /Page\n/Parent 2 0 R\n");
  pdf->AddLine("/MediaBox [0 0 612 792]\n");
  pdf->AddLine("/Contents 4 0 R\n");
  pdf->AddLine("/Annots [9 0 R 17 0 R]\n");
  pdf->AddLine("/Resources << /Font << /F1 5 0 R >> /ProcSet 6 0 R >>\n");
  pdf->CloseDict();
  pdf->CloseObject();

  /* Write a text content object, No. 4 */
  pdf->MakeTextContent(4, text1);

  /* Write a /Font Object, No. 5 */
  pdf->MakeFont(5, 1, "Helvetica");

  /* Write a /ProcSet object, No. 6 */
  pdf->StartObject(6);
  pdf->AddLine("[/PDF /Text]\n");
  pdf->CloseObject();

  /* Write a Gliederungsobjekt(/Outlines), No. 7 */
  pdf->StartObject(7);
  pdf->StartDict();
  pdf->AddLine("/Type /Outlines\n/Count 2\n/First 21 0 R /Last 22 0 R >>\n");
  pdf->CloseDict();
  pdf->CloseObject();

  /* Outline object 21 */
  pdf->MakeOutline(21, 7, 22, 3, "Seite 1");
  /* Outline object 22 */
  pdf->MakeOutline(22, 7, 0, 12, "Seite 2");

  /* schreibt ein /Info Object, No. 8 */
  pdf->MakeInfo(8, "E.Huckert", "Program pdfg", "1.3.1996");

  /* Write an annotation object, No. 9 */
  pdf->MakeAnnotation(9, 500, 400, 700, 500,
                      "Dies ist eine Annotation");

  /* Write a Link object (link to an ext. file) , object No. 17 */
  /* the external file called yy.pdf must be supplied           */
  //pdf->MakeLinkAnnotation(17, 12, 550, 550, 590, 610);
  pdf->MakeExtLinkAnnotation(17, "yy.pdf", 550, 550, 590, 610);

  /* Write a second /Page object, No.12 */
  pdf->StartObject(12);
  pdf->StartDict();
  pdf->AddLine("/Type /Page\n");
  pdf->AddLine("/Parent 2 0 R\n");
  pdf->AddLine("/MediaBox [0 0 612 792]\n");
  pdf->AddLine("/Contents 10 0 R\n");
  pdf->AddLine("/Resources <<\n/Font << /F1 5 0 R >>\n");
  pdf->AddLine("/ProcSet 13 0 R\n");
  pdf->AddLine("/XObject << /bild1 14 0 R >>\n>>\n");
  pdf->CloseDict();
  pdf->CloseObject();

  /* Write a Text Inhaltsobjekt No. 10 */
  /* Befehl zum Zeichnen Bild einfuegen     */
  //pdf->MakeTextContent(10, text2);
  pdf->StartObject(10);
  sprintf(ibuf, "250 300 m\nq 200 0 0 250 20 30 cm\n/bild1 Do\nQ\n");
  sprintf(hbuf, "<< /Length %d >>\nstream\n", 
          strlen(text2) + strlen(ibuf));
  pdf->AddLine(hbuf);
  pdf->AddLine(text2);
  pdf->AddLine(ibuf);
  pdf->AddLine("endstream\n");
  pdf->CloseObject();

  /* Write a /ProcSet object, No. 13 */
  pdf->StartObject(13);
  pdf->AddLine("[/PDF /Text /ImageB]\n");
  pdf->CloseObject();

  /* Just for test: we generate here a binary bitmap file */
  ofd = creat("test.bmp", CREAT_MASK);
  if (ofd > 0)
  {
    int x, y;
    char scanline[32];
    for (x=0; x < 32; x++)
      scanline[x] = 0x55;  // arbitrary pattern, here 01010101
    for (y=0; y < 128; y++)
    {
      write(ofd, scanline, sizeof(scanline));  
    }
    close(ofd);
  }

  /* Write an image object + a corresponding length object     */
  /* objects 14 und 15                                         */
  /* Note: the file test.bmp is not a Windows bitmap: it just  */
  /* contains binary data representing the bitmap (no header)! */
  /* Size of the bitmap: 256 x 128 x 1                         */
  pdf->MakeImage(14, "test.bmp", 32 * 8, 128, 1);

  // write now the root object
  // Note: we want that the second page should be opend at start!!!
  pdf->MakeRootObject(1, 2, 7, 12);

  delete pdf;
  return 0;
}   /* end main */
